import gym
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
import matplotlib.pyplot as plt
import os

# Define a simple neural network as the policy
class PolicyNetwork(nn.Module):
    def __init__(self, input_size, output_size, hidden_size=128):
        super(PolicyNetwork, self).__init__()
        self.fc1 = nn.Linear(input_size, hidden_size)
        self.relu = nn.ReLU()
        self.fc2 = nn.Linear(hidden_size, output_size)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x):
        x = self.relu(self.fc1(x))
        x = self.fc2(x)
        return self.softmax(x)

# Function to compute discounted returns
def compute_returns(rewards, gamma=0.99):
    returns = []
    R = 0
    for r in reversed(rewards):
        R = r + gamma * R
        returns.insert(0, R)
    returns = torch.tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + 1e-8)  # Normalize
    return returns

# Function to train the policy network using REINFORCE algorithm
def reinforce(env_name="CartPole-v1", num_episodes=500, gamma=0.99, learning_rate=5e-3):
    # Initialize environment and policy network
    env = gym.make(env_name)
    policy = PolicyNetwork(env.observation_space.shape[0], env.action_space.n)
    optimizer = optim.Adam(policy.parameters(), lr=learning_rate)
    
    episode_rewards = []
    
    for episode in range(num_episodes):
        episode_reward = 0
        log_probs = []
        rewards = []
        
        state = env.reset()
        done = False

        while not done:
            state_array = state[0] if isinstance(state, tuple) else state  # Extract array representing the state
            state_tensor = torch.tensor(state_array, dtype=torch.float32).unsqueeze(0)  # Convert to tensor
            action_probs = policy(state_tensor)
            action = torch.multinomial(action_probs, num_samples=1).item()
            log_probs.append(torch.log(action_probs[0][action]))  # Access the first element of the tensor

            # Take action in the environment
            step_output = env.step(action)
            next_state, reward, terminated,truncated, info = step_output  # Unpack step output
            done = truncated or terminated

            rewards.append(reward)  # Collect rewards

            # Update the state for the next iteration
            state = next_state
            
            episode_reward += reward  # Accumulate rewards

        # Compute discounted returns
        returns = compute_returns(rewards, gamma)
        
        # Compute loss and update policy
        policy_loss = -torch.sum(torch.stack(log_probs) * returns)
        optimizer.zero_grad()
        policy_loss.backward()
        optimizer.step()
        
        episode_rewards.append(episode_reward)
        print(f"Episode {episode + 1}/{num_episodes}, Reward: {episode_reward}")
        
    env.close()
    
    return episode_rewards



if __name__ == "__main__":
    # Train the policy network
    episode_rewards = reinforce()

    # Plotting the total reward across episodes
    plt.plot(np.arange(len(episode_rewards)), episode_rewards)
    plt.xlabel('Episode')
    plt.ylabel('Total Reward')
    plt.title('Total Reward Across Episodes')

    # Save the plot in a file named 'plots'
    plots_dir = 'plots'
    if not os.path.exists(plots_dir):
        os.makedirs(plots_dir)
    plot_file = os.path.join(plots_dir, 'reward_plot1.png')
    plt.savefig(plot_file)

    plt.show()
